/*-----------------------------------------------------------------------------
// Project            	GeometryKit
// Class              	G3DDefs
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The GeometryKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DDefs.h,v 1.1 2002/10/12 17:24:32 probert Exp $
//
//---------------------------------------------------------------------------*/

#ifndef __G3DDefs_h_INCLUDE
#define __G3DDefs_h_INCLUDE

#include <limits.h>
#include <math.h>

/******************************************************************************
 *
 * Enums
 *
 *****************************************************************************/

typedef enum {
  G3D_SCALE = 0,
  G3D_ROTATION,
  G3D_TRANSLATION,
  G3D_ID,
  G3D_GENERIC
} G3D_MatrixType;

/******************************************************************************
 *
 * Miscellaneous defines 
 *
 *****************************************************************************/

#define DEG2RAD  (PI/180.0f)
#define RAD2DEG  (180.0f/PI)

#ifndef __G3DIE__
#define __G3DIE__ __inline__ extern
#endif

#ifndef __G3DI__
#define __G3DI__ __inline__
#endif

#ifndef __G3DE__
#define __G3DE__ extern
#endif

#ifndef CFLOAT
#define CFLOAT const float
#endif

#ifndef CDOUBLE
#define CDOUBLE const double
#endif

/******************************************************************************
 *
 * Macros with 1 argument as defined i.e. in Graphics Gem II
 *
 *****************************************************************************/

#ifndef ABS
#define ABS(a) (((a)<0) ? -(a) : (a))
#endif

#ifndef FLOOR
#define FLOOR(a) (((a)<0) ? (int)(a) : -(int)(-a))
#endif

#ifndef CEIL
#define CEIL(a) (((a)==(int)(a)) ? (a) : (a)>0 ? 1+(int)(a) : -(1+ (int)(-a)))
#endif

#ifndef ROUND
#define ROUND(a) (((a)>0) ? (int)(a+0.5) : -(int)(0.5-a))
#endif

#ifndef ZSGN
#define ZSGN(a) (((a)<0) ? -1 : ((a)>0) ? 1 : 0)
#endif

#ifndef SGN
#define SGN(a) (((a)<0) ? -1 : 0)
#endif

#ifndef SQR
#define SQR(a) ((a)*(a))
#endif

/******************************************************************************
 *
 * Macros with 2 argument as defined i.e. in Graphics Gem II
 *
 *****************************************************************************/

#ifndef MIN
#define MIN(a,b) (((a)<(b)) ? (a) : (b)) 
#endif

#ifndef MAX
#define MAX(a,b) (((a)>(b)) ? (a) : (b)) 
#endif

#ifndef SWAP
#define SWAP(a,b) { a^=b; b^=a; a^=b} 
#endif

/******************************************************************************
 *
 * Linear interpolation between l and h
 *
 *****************************************************************************/

#ifndef LERP
#define LERP(v,l,h) ((l) + (((h)-(l)) * (v))) 
#endif

/******************************************************************************
 *
 * Clamp to range [l,h]
 *
 *****************************************************************************/

#ifndef CLAMP
#define CLAMP(v,l,h) ((v)<(l) ? (l) : (v)>(h) ? (h) : (v)) 
#endif

/******************************************************************************
 *
 * Mathematical Constants
 *
 *****************************************************************************/

#ifndef PI
#define PI  3.14159265358979323846
#endif

#ifndef TWOPI
#define TWOPI  6.283185306f
#endif

#ifndef PIOVER2
#define PIOVER2  1.570796327f
#endif

#ifndef E
#define E  2.718282f
#endif

#ifndef SQRT2
#define SQRT2  1.414214f
#endif

#ifndef SQRT3
#define SQRT3  1.732051f
#endif

#ifndef G3DEPSILON
#define G3DEPSILON 1.0e-8
#endif

#ifndef G3DBIGEPSILON
#define G3DBIGEPSILON 0.00001
#endif

#endif
