//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DMatrix4d
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The GeometryKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DMatrix4d.h,v 1.1 2002/10/12 17:24:32 probert Exp $
//-----------------------------------------------------------------------------

#ifndef __G3DMatrix4d_h_INCLUDE
#define __G3DMatrix4d_h_INCLUDE

#import <Foundation/Foundation.h>
#import <GeometryKit/G3DDefs.h>

@class G3DMatrix3d;
@class G3DTuple3d;
@class G3DTuple4d;
@class G3DVector3d;
@class G3DVector4d;
@class G3DQuaterniond;

@interface G3DMatrix4d : NSObject <NSCoding, NSCopying>
{
  @public

  double matrix[16];

  @protected

  int _mode;           // G3D_MatrixType

  double _t[16];
  double _i[16];
}

//-----------------------------------------------------------------------------
// init and free
//-----------------------------------------------------------------------------

- (id)init;
- (id)initWithMatrix4d:(G3DMatrix4d *)aMatrix;
- (id)initWithElements:(double *)vals;

- (id)initWithRotation:(G3DMatrix4d *)vals translation:(G3DVector3d *)t scale:(double)factor;
// self = S * R * T

- (id)initWithQuaternion:(G3DQuaterniond *)q translation:(G3DVector3d *)t scale:(double)s;

- (void)dealloc;

//-----------------------------------------------------------------------------
// math
//-----------------------------------------------------------------------------

- (void)addMatrix:(G3DMatrix4d *)a;
- (void)subMatrix:(G3DMatrix4d *)a;

- (void)multiplyByMatrix:(G3DMatrix4d *)a;
- (void)multiplyByMatrix:(G3DMatrix4d *)a and:(G3DMatrix4d *)b;
- (void)multiplyByMatrix:(G3DMatrix4d *)a andTransposed:(G3DMatrix4d *)b;
- (void)multiplyByTransposedMatrix:(G3DMatrix4d *)a and:(G3DMatrix4d *)b;
- (void)multiplyByTransposedMatrix:(G3DMatrix4d *)a andTransposed:(G3DMatrix4d *)b;

- (void)multiplyByMatrices:(NSArray *)matrices;

- (BOOL)invert;
- (G3DMatrix4d *)invertedMatrix;

- (void)transpose;
- (G3DMatrix4d *)transposedMatrix;

- (void)addScalar:(double)scalar;
- (void)subScalar:(double)scalar;

- (void)multiplyByScalar:(double)scalar;
- (void)divideByScalar:(double)scalar;

- (void)multiplyRow:(int)pos byValue:(double)val;
- (void)multiplyCol:(int)pos byValue:(double)val;

- (void)addElements:(double *)vals toRow:(int)rowPos;
- (void)addElements:(double *)vals toColumn:(int)colPos;

- (void)subElements:(double *)vals fromRow:(int)rowPos;
- (void)subElements:(double *)vals fromColumn:(int)colPos;

- (double)determinant;

- (G3DVector4d *)vectorByPreMultiplying:(G3DVector4d *)aTuple;
  // Returns aTuple * self;

- (G3DVector4d *)vectorByPostMultiplying:(G3DVector4d *)aTuple;
  // Returns self * aTuple

//-----------------------------------------------------------------------------
// accessor methods
//-----------------------------------------------------------------------------

- (G3DVector4d *)translation;
- (void)setTranslation:(G3DVector4d *)t;

- (G3DVector4d *)scale;
- (void)setScale:(G3DVector4d *)scale;

- (G3DQuaterniond *)quaternion;
- (void)setQuaternion:(G3DQuaterniond *)aQuat;

- (void)setMatrix:(G3DMatrix4d *)aMatrix;

- (void)setXRotation:(double)angle;
- (void)setYRotation:(double)angle;
- (void)setZRotation:(double)angle;
- (void)setRotation:(double)angle axis:(double *)axis;
- (void)setRotation:(double)angle axisVector:(G3DVector3d *)vec;

- (void)setRotation:(G3DMatrix4d *)vals translation:(G3DVector3d *)t scale:(double)factor;
- (void)setQuaternion:(G3DQuaterniond *)q translation:(G3DVector3d *)t scale:(double)s;

- (void)setElements:(double *)vals;
- (double *)elements;

- (G3DVector4d *)rowVectorAtIndex:(int)row;
- (G3DVector4d *)columnVectorAtIndex:(int)col;

- (void)setElement:(double)val atRow:(int)row column:(int)col;
- (double)elementAtRow:(int)row column:(int)col;

//-----------------------------------------------------------------------------
// misc
//-----------------------------------------------------------------------------

- (BOOL)isEqualToMatrix:(G3DMatrix4d *)aMatrix;

- (NSString *)description;

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder;
- (id)initWithCoder:(NSCoder *)aCoder;

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone;

@end

extern NSString *G3DMatrix4dException;

#endif




