//-----------------------------------------------------------------------------
// Project            	GeometryKit
// Class		G3DQuaterniond
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The GeometryKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DQuaterniond.h,v 1.1 2002/10/12 17:24:32 probert Exp $
//-----------------------------------------------------------------------------

/*
  Quaternion routines are partially based on the paper from
  Nick Bobick, GameDeveloper Vol.2, Issue 26. They are not entirely correct, 
  though, so I fixed them...
 */

#ifndef __G3DQuaterniond_h_INCLUDE
#define __G3DQuaterniond_h_INCLUDE

#import <Foundation/Foundation.h>
#import <GeometryKit/G3DTuple4d.h>

@class G3DMatrix3d;
@class G3DMatrix4d;
@class G3DVector3d;

@interface G3DQuaterniond : G3DTuple4d
{
}

//-----------------------------------------------------------------------------
// init and free
//-----------------------------------------------------------------------------

- (id)init;
  // Returns an unity quaternion, Designated initialiser!

- (id)initWithElements:(double *)values;
- (id)initWithQuaternion:(G3DQuaterniond *)aTuple;
    // these methods invoke the corresponding methods from G3DTuple4f!

- (id)initWithMatrix4d:(G3DMatrix4d *)aMatrix;
  // Matrix is a rotation matrix

- (id)initWithEulerRep:(double *)ev;

//-----------------------------------------------------------------------------
// Math
//-----------------------------------------------------------------------------

- (double)norm;
// N = sum(a_i^2)

- (void)conjugate;
// q* = [-v,w]

- (G3DQuaterniond *)conjugatedQuaternion;

- (void)invert;
// q^-1 = q*/N(q)

- (G3DQuaterniond *)invertedQuaternion;

- (void)normalise;
// q2 = q/L(q)

- (G3DQuaterniond *)normalisedQuaternion;

- (void)interpolate:(G3DQuaterniond *)quat factor:(double)val;
// Interpolates (SLERP - spherical linear interpolation) between self and quat

- (void)multiplyByQuaternion:(G3DQuaterniond *)a;
- (void)multiplyByQuaternion:(G3DQuaterniond *)a and:(G3DQuaterniond *)b;

- (void)multiplyByInvertedQuaternion:(G3DQuaterniond *)a;
- (void)multiplyByQuaternion:(G3DQuaterniond *)a andInverted:(G3DQuaterniond *)b;

- (void)multiplyByScalar:(double)value;

- (void)addQuaternion:(G3DQuaterniond *)aQuat;
- (void)subQuaternion:(G3DQuaterniond *)aQuat;

- (void)rotateByAngle:(double)angle axis:(G3DVector3d *)aVec;
// Builds a new quaternion q, multiplies self with q and normalises itself

//-----------------------------------------------------------------------------
// Accessor methods
//-----------------------------------------------------------------------------

- (void)setQuaternionWithMatrix4d:(G3DMatrix4d *)m;

- (G3DMatrix4d *)rotationMatrix4d;
// Returns a 4x4 rotation matrix, based upon self

- (G3DMatrix3d *)rotationMatrix3d;
// Returns a 3x3 rotation matrix, based upon self

- (void)setQuaternion:(G3DQuaterniond *)q;
// Invokes [super setValuesWithTuple:q]

- (void)setVector:(G3DVector3d *)vec angle:(double)val;
// Initialises the quaternion from an angle val and an axis vec 

- (G3DTuple4d *)angleAxisRepresentation;
// Returns a vector containing the axis in the first 3 values and the angle in the forth

@end

extern NSString *G3DQuaterniondException;

#endif





