//-----------------------------------------------------------------------------
// Project            	GeometryKit
// Class		G3DTrigonometry
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The GeometryKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DTrigonometry.m,v 1.1 2002/10/12 17:24:32 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DTrigonometry.h"
#import "G3DDefs.h"

NSString *TableAllocationException = @"TableAllocationException";

@implementation G3DTrigonometry

static float *costable;
static float *sintable;

//-----------------------------------------------------------------------------
// class methods
//-----------------------------------------------------------------------------

+ (void)initialize
{
  static BOOL tooLate = NO;

  if (!tooLate) {
    int	i;
    double r = 0;

    tooLate = YES;

    costable = (float *)calloc(4, 3600);
    if(costable == NULL) {
      [NSException raise:TableAllocationException format:@"Could not allocate the cosine table!"];
      return;
    }
    
    sintable = (float *)calloc(4, 3600);
    if(sintable == NULL) {
      [NSException raise:TableAllocationException format:@"Could not allocate the sine table!"];
      return;
    }

	
    for( i = 0; i < 3600; i++) {
      costable[i] = cos(r);
      r += .01;
    }
    
    r = 0;
    for( i = 0; i < 3600; i++) {
      sintable[i] = sin(r);
      r += .01;
    }
  }
}

+ (float)fastSin:(int)angle
{
  return(sintable[abs(angle % 3600)]);
}

+ (float)fastCos:(int)angle
{
  return(costable[abs(angle % 3600)]);
}

@end






