#import <Foundation/Foundation.h>
#import <Foundation/NSDebug.h>
#import <GeometryKit/G3DDefs.h>
#import <GeometryKit/G3DFunctions.h>
#import <GeometryKit/G3DBezier.h>
#import <GeometryKit/G3DVector3f.h>

#define START_TIMER sTime = [NSDate date]
#define END_TIMER eTime = [NSDate date]
#define PRINT_TIMER(str) printf("  %-20s\t %6.3f \t %6.3f\n", str, \
			[eTime timeIntervalSinceDate: sTime], \
			[eTime timeIntervalSinceDate: sTime]/baseline)

NSDate	*sTime = nil;
NSDate	*eTime = nil;
/* Set to a baseline to null out speed of runtime */
NSTimeInterval baseline = 0.0;


int main()
{
  id pool;

  pool = [[NSAutoreleasePool alloc] init];

  {
    // Three vectors to define the bezier curve
    //int i;
    G3DBezier* bezier = nil;
    NSMutableArray* controlPoints = [NSMutableArray arrayWithCapacity:3];
    NSMutableArray* controlPoints2 = [NSMutableArray arrayWithCapacity:3];
    float vectorElems1[3] = {0.0,0.0,0.0};
    float vectorElems2[3] = {1.0,0.0,0.0};
    float vectorElems3[3] = {0.0,1.0,0.0};
    float vectorElems4[3] = {0.0,1.0,1.0};
    
    G3DVector3f* pointOnBezier = nil;
    G3DVector3f *v1 = [[G3DVector3f alloc] initWithElements:vectorElems1];
    G3DVector3f *v2 = [[G3DVector3f alloc] initWithElements:vectorElems2];
    G3DVector3f *v3 = [[G3DVector3f alloc] initWithElements:vectorElems3];
    G3DVector3f *v4 = [[G3DVector3f alloc] initWithElements:vectorElems4];
    
    [controlPoints addObject:v1];
    [controlPoints addObject:v2];
    [controlPoints addObject:v3];
    
    [controlPoints2 addObjectsFromArray:controlPoints];

    NSLog(@"G3DBezier Test Tool");
    
    NSLog([v1 description]);
    NSLog([v2 description]);
    NSLog([v3 description]);
    NSLog([v4 description]);

    NSLog([controlPoints description]);
    NSLog([controlPoints2 description]);
    
    if ( [controlPoints isEqualToArray:controlPoints2] == YES ) {
      NSLog( @"Right: arrays are equal" );
    }
    else {
	NSLog( @"Bad: arrays should be equal" );
    }
    
    [controlPoints2 replaceObjectAtIndex:2 withObject:v4];
    if ( [controlPoints isEqualToArray:controlPoints2] == YES ) {
      NSLog( @"Bad: arrays are still equal" );
    }
    else {
      NSLog( @"Right: arrays are not equal anymore" );
    }
    
    bezier = [[G3DBezier alloc] initWithControlPoints:controlPoints];
    NSLog([bezier description]);
    
    //   START_TIMER;
    // 		for ( i = 0; i < 10000; i++ )
    // 		{
    pointOnBezier = [bezier pointAtParameter:0.5];
    NSLog([pointOnBezier description]);
    
    pointOnBezier = [bezier pointAtParameter:0.0];
    NSLog([pointOnBezier description]);
    
    pointOnBezier = [bezier pointAtParameter:1.0];
    NSLog([pointOnBezier description]);
    
    pointOnBezier = [bezier pointAtParameter:0.75];
    NSLog([pointOnBezier description]);
    
    pointOnBezier = [bezier pointAtParameter:0.25];
    NSLog([pointOnBezier description]);
    // 		}
    //   END_TIMER;
    //   PRINT_TIMER("Elapsed time during test");
    
    
    NSLog(@"Done...");
    
    [v1 autorelease];
    [v2 autorelease];
    [v3 autorelease];
  }
  
  [pool release];
  exit(0);
}
