#import <Foundation/Foundation.h>
#import <GeometryKit/G3DMatrix3d.h>
#import <GeometryKit/G3DFunctions.h>
#import <GeometryKit/G3DDefs.h>

NSString *G3DStringFromMatrix3(double *mat)
{
  return [NSString stringWithFormat: @"3x3 double matrix:\n\n%d %d %d \n%d %d %d \n%d %d %d \n",mat[0],mat[3],mat[6],mat[1],mat[4],mat[7],mat[2],mat[5],mat[8]];
}

int main()
{
  id pool;

  pool = [[NSAutoreleasePool alloc] init];

  {
    double vals[9] = {1,2,3,0,5,2,1,0,9};
    double vals2[9] = {1,0,0,0,2,0,0,0,3};
    double axis[3] = {1.0,0.0,1.0};

    double p[3] = {1.0, 0.0, 0.0};
    double a[3] = {0.0, 1.0, 0.0};
    double res[3];
    double mat[9];

    G3DMatrix3d *m;

    NSLog(@"G3DMatrix3d Test Tool");

    G3DMakeRotation3d(mat, 0.5, a);
    G3DVector3dXMatrix3d(res,  p, mat);

    NSLog(@"Rotation matrix: %@",G3DStringFromMatrix3(mat));
    NSLog(@"Vector rotated : %f %f %f", res[0], res[1], res[2]);

    m = [[G3DMatrix3d alloc] initWithElements:vals];

    // First matrix...
    NSLog([m description]);
    NSLog([NSString stringWithFormat:@"Determinante: %f",[m determinant]]);
    [m transpose];
    NSLog(@"Transposed matrix:");
    NSLog([m description]);

    // Second matrix...
    [m setElements:(double*)G3DIdentityMatrix3d];
    NSLog([m description]);
    NSLog([NSString stringWithFormat:@"Determinante: %f",[m determinant]]);
    if(![m invert]) {
      NSLog(@"Couldn't invert matrix");
    }
    else {
      NSLog(@"Inverted matrix:");
      NSLog([m description]);
    }

    // Third matrix...
    [m setElements:vals2];
    NSLog([m description]);
    NSLog([NSString stringWithFormat:@"Determinante: %f",[m determinant]]);

    if(![m invert]) {
      NSLog(@"Couldn't invert matrix");
    }
    else {
      NSLog(@"Inverted matrix:");
      NSLog([m description]);
    }

    // Column operation
    NSLog(@"Subtract 1,0,1 from column 2");
    [m subElements:axis fromColumn:1];
    NSLog([m description]);

    // Column operation
    NSLog(@"Column 2");
    NSLog([[m columnVectorAtIndex:1] description]);

    // Row operation
    NSLog(@"Add 1,0,1 to row 2");
    [m addElements:axis toRow:1];
    NSLog([m description]);

    // Row operation
    NSLog(@"Row 2");
    NSLog([[m rowVectorAtIndex:1] description]);

    // Some elements?!...
    NSLog([NSString stringWithFormat:@"Element 2,3: %f",[m elementAtRow:1 column:2]]);
    NSLog([NSString stringWithFormat:@"Element 3,2: %f",[m elementAtRow:2 column:1]]);

    // Make a rotation matrix...
    NSLog(@"Rotation with angle 45 and axis 1,0,1");
    [m setRotation:45.5 axis:axis];
    NSLog([m description]);
    NSLog([NSString stringWithFormat:@"Element 1,3: %f",[m elementAtRow:0 column:2]]);

    // Make a rotation matrix...
    NSLog(@"Rotation with angle 33.3 and axis 1,0,0");
    [m setXRotation:33.333];
    NSLog([m description]);
    // Make a rotation matrix...
    NSLog(@"Rotation with angle 33.3 and axis 0,1,0");
    [m setYRotation:33.333];
    NSLog([m description]);
    // Make a rotation matrix...
    NSLog(@"Rotation with angle 33.3 and axis 0,0,1");
    [m setZRotation:33.333];
    NSLog([m description]);

    NSLog(@"Done...");

    [m release];
  }
  
  [pool release];
  exit(0);
}
